from .fixtures import *  # required for all
import time
import uuid
import yaml
from contextlib import contextmanager
from ..repo import Repo
import os
import subprocess
import click
import inspect
import os
from pathlib import Path
from .tools import gimera_apply
from .tools import _make_remote_repo
from .tools import clone_and_commit

from ..consts import gitcmd as git

def test_basicbehaviour(temppath):
    """
    * put same repo integrated and submodule into main repo
    * add file2.txt on remote
    * check after apply that file exists in both
    * make a patch in integrated version
    """
    workspace = temppath / "workspace"

    remote_main_repo = _make_remote_repo(temppath / "mainrepo")
    remote_sub_repo = _make_remote_repo(temppath / "sub1")

    subprocess.check_output(
        git + ["clone", "file://" + str(remote_main_repo), workspace.name],
        cwd=workspace.parent,
    )
    os.environ["GIMERA_NON_INTERACTIVE"] = "1"

    # region gimera config
    repos = {
        "repos": [
            {
                "url": f"file://{remote_sub_repo}",
                "branch": "branch1",
                "path": "submodules/sub1",
                "patches": [],
                "type": "submodule",
            },
            {
                "url": f"file://{remote_sub_repo}",
                "branch": "branch1",
                "path": "integrated/sub1",
                "patches": ["integrated/sub1_patches"],
                "type": "integrated",
            },
        ]
    }
    # endregion

    (workspace / "gimera.yml").write_text(yaml.dump(repos))
    (workspace / "main.txt").write_text("main repo")
    subprocess.check_call(git + ["add", "main.txt"], cwd=workspace)
    subprocess.check_call(git + ["add", "gimera.yml"], cwd=workspace)
    subprocess.check_call(git + ["commit", "-am", "on main"], cwd=workspace)
    subprocess.check_call(git + ["push"], cwd=workspace)
    (workspace / repos["repos"][1]["patches"][0]).mkdir(exist_ok=True, parents=True)
    os.chdir(workspace)
    os.environ["GIMERA_NON_THREADED"] = "1"
    gimera_apply([], None)
    subprocess.check_call(git + ["add", "gimera.yml"], cwd=workspace)
    assert not Repo(workspace).staged_files
    # subprocess.check_call(git + ["commit", "-am", "updated gimera"], cwd=workspace)

    click.secho(
        "Now we have a repo with two subrepos; now we update the subrepos and pull"
    )
    with clone_and_commit(remote_sub_repo, "branch1") as repopath:
        (repopath / "file2.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file2.txt"], cwd=repopath)
        subprocess.check_call(git + ["commit", "-am", "file2 added"], cwd=repopath)

    os.chdir(workspace)
    os.environ["GIMERA_NON_INTERACTIVE"] = "1"
    gimera_apply([], update=True)

    click.secho(str(workspace), fg="green")
    assert (workspace / "submodules" / "sub1" / "file2.txt").exists()
    assert (workspace / "integrated" / "sub1" / "file2.txt").exists()

    # check dirty - disabled because the command is_path_dirty is not cool
    (workspace / "integrated" / "sub1" / "file2.txt").write_text("a change!")
    (workspace / "integrated" / "sub1" / "file3.txt").write_text("a new file!")
    (workspace / "file4.txt").write_text(
        "a new file!"
    )  # should not stop the process but should not be committed

    # annotation: it would be bad if file3.txt would be gone
    # and also the change of file2.txt

    # now lets make a patch for new integrated/sub1/file3.txt and changed integrated/sub1/file2.txt
    os.chdir(workspace)
    os.environ["GIMERA_EXCEPTION_THAN_SYSEXIT"] = "1"
    (workspace / repos["repos"][1]["patches"][0]).mkdir(exist_ok=True, parents=True)
    gimera_apply([], update=True)
    assert (workspace / "integrated" / "sub1" / "file3.txt").exists()

    # now lets make an update and see if patches are applied
    with clone_and_commit(remote_sub_repo, "branch1") as repopath:
        (repopath / "file5.txt").write_text("This is a new function")
        subprocess.check_call(git + ["add", "file5.txt"], cwd=repopath)
        subprocess.check_call(git + ["commit", "-am", "file5 added"], cwd=repopath)

    # should apply patches now
    os.chdir(workspace)
    gimera_apply([], update=False)

    # check if test is applied
    assert "file4.txt" in [x.name for x in Repo(workspace).untracked_files]
    assert (workspace / "integrated" / "sub1" / "file3.txt").exists()
    assert (
        "a change"
        in (
            (workspace / "integrated" / "sub1" / "file3.txt").parent / "file2.txt"
        ).read_text()
    )

    # now lets edit that patch again
    Repo(workspace).simple_commit_all()
    patchfile = list((workspace / "integrated" / "sub1_patches").glob("*"))[
        0
    ].relative_to(workspace)
    os.chdir(workspace)
    from ..patches import _edit_patch as edit_patch

    edit_patch([patchfile])
    dirty_files = Repo(workspace).all_dirty_files_absolute
    assert (workspace / str(patchfile)) not in dirty_files
    assert (workspace / "integrated/sub1/file3.txt") in dirty_files
    assert (workspace / "integrated/sub1/file2.txt") in dirty_files




from .fixtures import * # required for all
import os
import yaml
from pathlib import Path
import subprocess
from ..consts import gitcmd as git
from ..repo import Repo
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_apply

# test change something in submodule and check if gets lost if not pushed

def test_switch_submodule_to_integrated_dont_loose_changes(temppath):
    """
    A gimera sub has changes and is submodule.
    Changes shall not be lost when switching to integrated modus.
    """
    workspace = temppath / "test_switch_submodule_to_integrated_dont_loose_changes"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")

    repos_sub = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    repos_int = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "sub1",
                "patches": [],
                "type": "integrated",
            },
        ]
    }
    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    (workspace_main / "main.txt").write_text("main repo")
    repo = Repo(workspace_main)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))

    os.chdir(workspace_main)
    gimera_apply([], None)

    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_int))
    repo.simple_commit_all()

    # make it dirty
    dirty_file = workspace_main / "sub1" / "file1.txt"
    original_content = dirty_file.read_text()
    dirty_file.write_text("i changed the file")

    os.chdir(workspace_main)
    try:
        gimera_apply([], None, raise_exception=True)
    except Exception:
        pass
    else:
        raise NotImplementedError("Should warn about changed content.")

    # switch back
    dirty_file.write_text(original_content)
    gimera_apply([], None)

    # now change in the integrated mode the file and try to switch to submodule
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_sub))
    dirty_file.write_text("dirty content")
    try:
        gimera_apply([], None, raise_exception=True)
    except Exception:
        pass
    else:
        raise NotImplementedError("Should warn about changed content.")
    dirty_file.write_text(original_content)
    gimera_apply([], None)


def test_switch_submodule_to_integrated_dont_loose_changes_with_subsub_repos(temppath):
    """
    A gimera sub has changes and is submodule.
    Changes shall not be lost when switching to integrated modus.
    """
    workspace = (
        temppath
        / "test_switch_submodule_to_integrated_dont_loose_changes_with_subsub_repos"
    )
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")
    repo_sub = _make_remote_repo(temppath / "sub1")
    repo_subsub = _make_remote_repo(temppath / "subsub")

    def switch_to(gimera_path, ttype):
        content = yaml.safe_load(gimera_path.read_text())
        for repo in content["repos"]:
            repo["type"] = ttype
        gimera_path.write_text(yaml.dump(content))

    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "gimera.yml").write_text(
            yaml.dump(
                {
                    "repos": [
                        {
                            "url": f"file://{repo_subsub}",
                            "branch": "main",
                            "path": "subsub",
                            "patches": [],
                            "type": "integrated",
                        },
                    ]
                }
            )
        )
        os.chdir(repopath)
        gimera_apply([], None)

    repos_gimera = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "main",
                "path": "sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    with clone_and_commit(repo_sub, "main") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    os.chdir(workspace_main)
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_gimera))
    (workspace_main / "main.txt").write_text("main repo")
    repo = Repo(workspace_main)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))

    os.chdir(workspace_main)
    gimera_apply([], None, recursive=True, strict=True)

    dirty_file = workspace_main / "sub1" / "subsub" / "file1.txt"
    original_content = dirty_file.read_text()
    dirty_file.write_text("changed it")
    os.chdir(workspace_main / "sub1")
    switch_to(Path("gimera.yml"), "submodule")
    try:
        gimera_apply([], None, raise_exception=True)
    except Exception:
        pass
    else:
        raise ValueError("Error for diff losts expected.")

    dirty_file.write_text(original_content)
    gimera_apply([], None)
    dirty_file.write_text("dirty")
    os.chdir(workspace_main / "sub1")
    switch_to(Path("gimera.yml"), "integrated")
    try:
        gimera_apply([], None)
    except Exception:
        pass
    else:
        raise ValueError("Error for diff losts expected.")



#!/usr/bin/env python3
import atexit
import tempfile
import re
from contextlib import contextmanager
import os
from datetime import datetime
import inquirer
import click
import sys
from pathlib import Path
from .repo import Repo, Remote
from .gitcommands import GitCommands
from .tools import _raise_error
from .consts import REPO_TYPE_INT, REPO_TYPE_SUB
from .config import Config
from .patches import _get_available_patchfiles
from .tools import try_rm_tree
from .tools import _get_missing_repos
from .tools import _get_main_repo
from .apply import _apply
from .commit import _commit
from .patches import _edit_patch
from .tools import rmtree
from . import runtime_state


@click.group()
def cli():
    pass


def _expand_repos(repos):
    def unique():
        config = Config()
        for repo in repos:
            if "*" not in repo:
                if repo.endswith("/"):
                    repo = repo[:-1]
                yield repo
            repo = repo.replace("*", ".*")
            for candi in config.repos:
                if not candi.enabled:
                    continue
                if re.findall(repo, str(candi.path)):
                    yield str(candi.path)

    res = list(set(unique()))
    exact_match = [x for x in res if x in repos]
    if len(exact_match) == 1:
        return exact_match
    return res


@cli.command(name="clean", help="Removes all git-dirty items")
def clean():
    Cmd = GitCommands()
    if not Cmd.dirty:
        click.secho("Everything already clean!", fg="green")
        return
    Cmd.output_status()
    doit = inquirer.confirm(
        "Continue cleaning? All local changes are lost.", default=True
    )
    if not doit:
        return
    Repo(Cmd.path).full_clean()
    click.secho("Cleaning done.", fg="green")
    Cmd.output_status()


@cli.command(name="combine-patch", help="Combine patches")
def combine_patches():
    click.secho("\n\nHow to combine patches:\n", fg="yellow")
    click.secho("1. Please install patchutils:\n\n\tapt install patchutils\n")
    click.secho("2. combinediff patch1 patch2 > patch_combined\n")


def _get_available_repos(ctx, param, incomplete):
    repos = []

    os.system(f"echo '{incomplete}' >> /tmp/gim")

    if "/" in incomplete:
        incomplete = incomplete.replace("/", "/*") + "*"
    else:
        incomplete = "*" + incomplete + "*"

    repos = list(_expand_repos([incomplete]))
    os.system(f"echo '{repos}' >> /tmp/gim")
    repos = list(map(str, repos))

    # repos = list(map(lambda x: '/'.join(x.split("/")[1:]), repos))
    def remove_parts(x):
        count = incomplete.count("/")
        x = x.split("/")[count:]
        return '/'.join(x)
    repos = list(map(remove_parts, repos))

    prefix = "/".join(incomplete.split("/")[:-1])
    repos = list(map(lambda x: (prefix and prefix + "/" or "") + x, repos))
    return sorted(repos)


@cli.command(name="apply", help="Applies configuration from gimera.yml")
@click.argument("repos", nargs=-1, default=None, shell_complete=_get_available_repos)
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="If set, then latest versions are pulled from remotes.",
)
@click.option(
    "--all-integrated",
    is_flag=True,
    help="Overrides setting in gimera.yml and sets 'integrated' for all.",
)
@click.option(
    "-S",
    "--all-submodule",
    is_flag=True,
    help="Overrides setting in gimera.yml and sets 'submodule' for all.",
)
@click.option(
    "-s",
    "--strict",
    is_flag=True,
    help=(
        "If set, then submodules of 'integrated' branches are not automatically "
        "set to integrated. They stay how they are defined in the children gimera.yml files."
    ),
)
@click.option(
    "-r",
    "--recursive",
    is_flag=True,
    help=("Executes recursive gimeras (analog to submodules initialization)"),
)
@click.option(
    "-P",
    "--no-patches",
    is_flag=True,
)
@click.option(
    "-m",
    "--missing",
    is_flag=True,
    help="Only updates paths, that dont exist yet in filesystem",
)
@click.option(
    "--remove-invalid-branches",
    is_flag=True,
    help="If branch does not exist in repository, the configuration item is removed.",
)
@click.option(
    "-I",
    "--non-interactive",
    is_flag=True,
    help="",
)
@click.option(
    "-C",
    "--no-auto-commit",
    is_flag=True,
    help="",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="",
)
@click.option(
    "-n",
    "--no-fetch",
    is_flag=True,
    help="",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="",
)
@click.option(
    "-SHA",
    "--no-sha-update",
    is_flag=True,
    help="If set, then in gimera.yml the shas are not updated.",
)
@click.option(
    "-M",
    "--migrate-changes",
    is_flag=True,
    help="Keeps changes to repo",
)
@click.option(
    "--raise-exception",
    is_flag=True,
    help="Raises exception instead of exit of program",
)
@click.option(
    "--do-not-apply-patches",
    is_flag=True,
    help="Do not apply patches",
)
@click.option(
    "--run-pre-commits",
    is_flag=True,
    help="Run pre-commit hooks at the end of apply",
)
@click.option(
    "--clear-cache", is_flag=True,
)
@click.option(
    "--clear-zip-cache", is_flag=True,
)
@click.option(
    "--no-cache", is_flag=True,
)
def apply(
    repos,
    update,
    all_integrated,
    all_submodule,
    strict,
    recursive,
    no_patches,
    missing,
    remove_invalid_branches,
    non_interactive,
    no_auto_commit,
    force,
    no_fetch,
    verbose,
    no_sha_update,
    migrate_changes,
    raise_exception,
    do_not_apply_patches,
    run_pre_commits,
    clear_cache,
    clear_zip_cache,
    no_cache,
):
    import shutil

    if not run_pre_commits:
        os.environ["GIMERA_NO_PRECOMMIT"] = "1"
    if verbose:
        os.environ["GIMERA_VERBOSE"] = "1"
    if no_sha_update:
        os.environ["GIMERA_NO_SHA_UPDATE"] = "1"
    if force:
        os.environ["GIMERA_FORCE"] = "1"
    if non_interactive:
        os.environ["GIMERA_NON_INTERACTIVE"] = "1"
        os.environ["GIT_TERMINAL_PROMPT"] = "0"
    if do_not_apply_patches:
        os.environ['GIMERA_DO_NOT_APPLY_PATCHES'] = "1"
    if all_integrated and all_submodule:
        _raise_error("Please set either -I or -S")
    if clear_cache:
        os.environ['GIMERA_CLEAR_CACHE'] = "1"
    if clear_zip_cache:
        os.environ['GIMERA_CLEAR_ZIP_CACHE'] = "1"
    if no_cache:
        os.environ['GIMERA_NO_CACHE'] = "1"
    ttype = None
    ttype = REPO_TYPE_INT if all_integrated else ttype
    ttype = REPO_TYPE_SUB if all_submodule else ttype

    repos = list(_expand_repos(repos))

    if missing:
        config = Config()
        repos = list(map(lambda x: str(x.path), _get_missing_repos(config)))
        if not repos:
            click.secho("Nothing to do - all paths exist.", fg="green")
            return

    try:
        _apply(
            repos,
            update,
            force_type=ttype,
            strict=strict,
            recursive=recursive,
            no_patches=no_patches,
            remove_invalid_branches=remove_invalid_branches,
            auto_commit=not no_auto_commit,
            no_fetch=no_fetch,
            migrate_changes=migrate_changes,
            raise_exception=raise_exception,
        )
    except Exception as ex:
        from . import snapshot

        snapshot.cleanup()
        raise

    if not run_pre_commits and repos and shutil.which("pre-commit"):
        dirs = " ".join(repos)
        click.secho(
            f"\nYou may run pre-commit for the applied repos:\n\n"
            f"  pre-commit run --files $(git ls-files {dirs})\n",
            fg="yellow",
        )


def clean_branch_names(arr):
    for x in arr:
        x = x.strip()
        if x.startswith("* "):
            x = x[2:]
        yield x


@cli.command()
@click.option(
    "-x",
    "--execute",
    is_flag=True,
    help=("Execute the script to insert completion into users rc-file."),
)
def completion(execute):
    shell = os.environ["SHELL"].split("/")[-1]
    rc_file = Path(os.path.expanduser(f"~/.{shell}rc"))
    line = f'eval "$(_GIMERA_COMPLETE={shell}_source gimera)"'
    if execute:
        content = rc_file.read_text().splitlines()
        if not list(
            filter(
                lambda x: line in x and not x.strip().startswith("#"),
                content,
            )
        ):
            content += [f"\n{line}"]
            click.secho(
                f"Inserted successfully\n{line}" "\n\nPlease restart you shell."
            )
            rc_file.write_text("\n".join(content))
        else:
            click.secho("Nothing done - already existed.")

    click.secho("\n\n" f"Insert into {rc_file}\n\n" f"echo 'line' >> {rc_file}" "\n\n")


@cli.command()
@click.option("-u", "--url", required=True)
@click.option("-b", "--branch", required=True)
@click.option("-p", "--path", required=True)
@click.option("-t", "--type", required=True)
def add(url, branch, path, type):
    data = {
        "url": url,
        "branch": branch,
        "path": path,
        "type": type,
    }
    config = Config()
    repos = config.repos
    for repo in repos:
        if str(repo.path) == path:
            config._store(repo, data)
            break
    else:
        ri = Config.RepoItem(config, data)
        config._store(ri, data)


@cli.command()
def check_all_submodules_initialized():
    if not _check_all_submodules_initialized():
        sys.exit(-1)


def _check_all_submodules_initialized():
    root = Path(os.getcwd())

    def _get_all_submodules(root):
        for submodule in Repo(root).get_submodules():
            path = submodule.path
            yield submodule
            if submodule._git_path.exists():
                yield from _get_all_submodules(path)

    error = False
    for submodule in _get_all_submodules(root):
        if not submodule._git_path.exists():
            click.secho(f"Not initialized: {submodule.path}", fg="red")
            error = True

    return not error


@cli.command()
@click.argument(
    "patchfiles", nargs=-1, shell_complete=_get_available_patchfiles, required=True
)
def edit_patch(patchfiles):
    _edit_patch(patchfiles)


@cli.command
def abort():
    for repo in Config().repos:
        if repo.edit_patchfile:
            repo.config._store(
                repo,
                {
                    "edit_patchfile": "",
                },
            )


@cli.command()
def status():
    config = Config()
    repos = list(_get_missing_repos(config))
    for repo in repos:
        click.secho(f"[{repo.type[0].upper()}] {repo.path}", fg="red")
    main_repo = _get_main_repo()
    for repo in config.get_repos(None):
        full_path = main_repo.path / repo.path
        if not full_path.exists():
            continue
        eff_S = bool(main_repo.is_path_a_submodule(repo.path))
        deviates = (
            repo.type == REPO_TYPE_INT
            and eff_S
            or repo.type == REPO_TYPE_SUB
            and not eff_S
        )
        text = f"[{repo.type[0].upper()}] {repo.path}"
        if deviates:
            text += " IS NOW " + ("submodule" if eff_S else "integrated")
        click.secho(text, fg="green" if not deviates else "yellow")


@cli.command(
    name="commit", help="Collects changes and commits them to the specified branch."
)
@click.argument(
    "repo", default=None, shell_complete=_get_available_repos, required=True
)
@click.argument("message", required=True)
@click.option(
    "-p",
    "--preview",
    is_flag=True,
)
@click.argument("branch", required=False)
def commit(repo, branch, message, preview):
    return _commit(repo, branch, message, preview)


@cli.command(help="Removes all dirty")
def purge():
    config = Config()
    repos = config.get_repos(None)
    for repo in repos:
        click.secho(f"Deleting: {repo.path}")
        try_rm_tree(repo.path)


@cli.command()
def list_snapshots():
    from .snapshot import list_snapshots

    main_repo = _get_main_repo()

    for snap in reversed(list(list_snapshots(main_repo.path))):
        click.secho(snap, fg="green")


@cli.command()
@click.argument("name", required=True)
@click.argument("repos", nargs=-1, default=None, shell_complete=_get_available_repos)
def snap(name, repos):
    from .snapshot import snapshot_recursive

    os.environ["GIMERA_TOKEN"] = datetime.now().strftime(f"%Y-%m-%d-%H%M%S-{name}")
    main_repo = _get_main_repo()
    if not repos:
        config = Config()
        repos = [x.path for x in config.get_repos(None) if x.path.exists()]
    else:
        repos = list(_expand_repos(repos))
    token = snapshot_recursive(
        main_repo.path, [main_repo.path / repo for repo in repos]
    )
    click.secho(f"Snapshot stored under token: {token}", fg="green")


@cli.command()
@click.argument("repos", nargs=-1, default=None, shell_complete=_get_available_repos)
def snaprestore(repos):
    from .snapshot import snapshot_restore
    from .snapshot import get_snapshots

    main_repo = _get_main_repo()
    snapshots = get_snapshots(main_repo.path)
    token = inquirer.prompt(
        [
            inquirer.List(
                "snapshot",
                "Choose snapshot",
                default=True,
                choices=snapshots,
            )
        ]
    )["snapshot"]
    filter_repos = []
    if repos:
        repos = list(_expand_repos(repos))
        for repo in repos:
            filter_repos.append(str(repo))

    snapshot_restore(main_repo.path, filter_repos, token=token)


def cleanup():
    for key, path in runtime_state['temppaths'].items():
        try:
            rmtree(path)
        except:
            click.secho(f"Could not wipe: {path}", fg='yellow')
atexit.register(cleanup)


from .fixtures import *  # required for all
import os
import yaml
import subprocess
from ..consts import gitcmd as git
from ..repo import Repo
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_apply
from ..tools import _make_sure_hidden_gimera_dir


def test_snapshot_and_restore_simple_add_delete_modify_direct_subrepo_submodule(
    temppath,
):
    repo_sub = _make_remote_repo(temppath / "sub1")
    repos_yaml = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "a/b/sub1",
                "patches": [],
                "type": "submodule",
            },
        ]
    }
    _test_snapshot_and_restore_simple_add_delete_modify_direct(
        temppath, repo_sub, repos_yaml
    )


def test_snapshot_and_restore_simple_add_delete_modify_direct_subrepo_integrated(
    temppath,
):
    repo_sub = _make_remote_repo(temppath / "sub1")
    repos_yaml = {
        "repos": [
            {
                "url": f"file://{repo_sub}",
                "branch": "branch1",
                "path": "a/b/sub1",
                "patches": [],
                "type": "integrated",
            },
        ]
    }
    _test_snapshot_and_restore_simple_add_delete_modify_direct(
        temppath, repo_sub, repos_yaml
    )


def _test_snapshot_and_restore_simple_add_delete_modify_direct(
    temppath, repo_sub, repos_yaml
):
    from ..snapshot import snapshot_recursive
    from ..snapshot import snapshot_restore

    workspace = temppath / "test_snapshot_and_restore"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")

    with clone_and_commit(repo_sub, "branch1") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_yaml))
    (workspace_main / "main.txt").write_text("main repo")
    repo = Repo(workspace_main)
    _make_sure_hidden_gimera_dir(repo.path)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))
    os.chdir(workspace_main)
    gimera_apply([], None)

    # make it dirty
    dirty_file = workspace_main / repos_yaml["repos"][0]["path"] / "file1.txt"
    original_content = dirty_file.read_text()
    dirty_file.write_text("i changed the file")
    # add a file
    added_file = workspace_main / repos_yaml["repos"][0]["path"] / "newfile.txt"
    added_file.write_text("new file")
    # delete a file
    deleted_file = workspace_main / repos_yaml["repos"][0]["path"] / "dont_look_at_me"
    deleted_file.unlink()

    os.chdir(workspace_main)
    snapshot_path = workspace_main / repos_yaml["repos"][0]["path"]
    snapshot_recursive(workspace_main, [snapshot_path])

    # reapply
    os.chdir(workspace_main)
    os.environ["GIMERA_FORCE"] = "1"
    gimera_apply([], None)

    assert dirty_file.read_text() == original_content

    # restore
    snapshot_restore(workspace_main, [snapshot_path])

    # make sure that situation is like before:
    assert dirty_file.read_text() == "i changed the file"
    assert added_file.exists()
    assert not deleted_file.exists()


def test_snapshot_two_root_submodules(temppath):
    """
    Challenge is, that the logic of snapshots can handle the root paths
    """
    workspace = temppath / "test_snapshot_two_root_submodules"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "main")
    repo_sub1 = _make_remote_repo(temppath / "sub1")
    repo_sub2 = _make_remote_repo(temppath / "sub2")

    with clone_and_commit(repo_sub1, "branch1") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()
    with clone_and_commit(repo_sub2, "branch2", ["-b"]) as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "repo_main.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        (repopath / "gimera.yml").write_text(
            yaml.dump(
                {
                    "repos": [
                        {
                            "url": str(repo_sub1),
                            "branch": "branch1",
                            "path": "sub1",
                            "type": "integrated",
                        },
                        {
                            "url": str(repo_sub2),
                            "branch": "branch2",
                            "path": "sub2",
                            "type": "integrated",
                        },
                    ]
                }
            )
        )
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    os.chdir(workspace_main)
    gimera_apply([], None)
    # assert everything is there
    assert (workspace_main / "sub1").exists()
    assert (workspace_main / "sub2").exists()

    # make it dirty
    dirty_file1 = workspace_main / "sub1" / "repo_sub.txt"
    dirty_file2 = workspace_main / "sub2" / "repo_sub.txt"
    assert dirty_file1.exists()
    assert dirty_file2.exists()
    dirty_file1.write_text("1")
    dirty_file2.write_text("2")
    os.environ["GIMERA_FORCE"] = "0"
    gimera_apply([], None, recursive=True, migrate_changes=True)

    assert dirty_file1.read_text() == "1"
    assert dirty_file2.read_text() == "2"



import os
from pathlib import Path
from .tools import safe_relative_to, yieldlist, X, wait_git_lock
from .consts import gitcmd as git


class GitCommands(object):
    def __init__(self, path=None):
        self.path = Path(path or os.getcwd())
        self.path_absolute = self.path.absolute()

    @property
    def configdir(self):
        from .repo import Repo

        stop_at = Repo(self.path_absolute).root_repo
        here = self.path_absolute
        while True:
            default = here / ".git"
            if default.exists() and default.is_dir():
                return default
            if default.is_file():
                path = default.read_text().strip().split("gitdir:")[1].strip()
                return (here / path).resolve()

            if here == stop_at:
                break
            here = here.parent
        raise Exception("Config dir not found")

    def X(self, *params, allow_error=False, env=None, output=None):
        if output is None:
            output = False
        with wait_git_lock(self.path_absolute):
            kwparams = {
                "output": output,
                "allow_error": allow_error,
                "env": env,
            }
            if self.path.exists():
                # case not existing at recreating cache dir e.g.
                kwparams["cwd"] = self.path
            return X(*params, **kwparams)

    def out(self, *params, allow_error=False, env=None):
        return X(*params, output=True, cwd=self.path, allow_error=allow_error, env=env)

    def _parse_git_status(self):
        for line in X(
            *(
                git
                + [
                    "status",
                    "--porcelain",
                    "--untracked-files=all",
                ]
            ),
            cwd=self.path_absolute,
            output=True,
        ).splitlines():
            # splits: A  asdas
            #         M   asdasd
            #          M  asdsad
            #         ??  asasdasd
            modifier = line[:2]
            path = line.strip().split(" ", 1)[1].strip()
            if path.startswith(".."):
                continue

            yield modifier, Path(path)

    @property
    @yieldlist
    def staged_files(self):
        for modifier, path in self._parse_git_status():
            if modifier[0] in ["A", "M", "D"]:
                yield path

    @property
    @yieldlist
    def dirty_existing_files(self):
        for modifier, path in self._parse_git_status():
            if modifier[0] == "M" or modifier[1] == "M" or modifier[1] == "D":
                yield path

    @property
    @yieldlist
    def all_dirty_files(self):
        return self.untracked_files + self.dirty_existing_files

    @property
    @yieldlist
    def all_dirty_files_absolute(self):
        res = self.untracked_files + self.dirty_existing_files
        res = list(map(lambda x: self.path_absolute / x, res))
        return res

    @property
    @yieldlist
    def untracked_files(self):
        for modifier, path in self._parse_git_status():
            if modifier == "??" or modifier[0] == "A":
                yield path

    @property
    @yieldlist
    def untracked_files_absolute(self):
        for file in self.untracked_files:
            yield self.path_absolute / file

    @property
    def dirty(self):
        return bool(list(self._parse_git_status()))

    def is_submodule(self, path):
        path = self._combine(path)
        for line in X(
            *(git + ["submodule", "status"]), output=True, cwd=self.path_absolute
        ).splitlines():
            line = line.strip()
            _, _path, _ = line.split(" ", 2)
            if _path == str(path.relative_to(self.path_absolute)):
                return path

    def _combine(self, path):
        """
        Makes a new path
        """
        path = self.path / path
        path.relative_to(self.path)
        return path

    def output_status(self):
        self.X(*(git + ["status"]))

    def get_all_branches(self):
        """
        4031c5eb19120f76a91b7cd9052bb27c5efe159a refs/heads/17.0
        2e45846285c6afc396a7bbadaa9dad54360ed51c refs/heads/main
        4031c5eb19120f76a91b7cd9052bb27c5efe159a refs/remotes/origin/17.0
        2e45846285c6afc396a7bbadaa9dad54360ed51c refs/remotes/origin/HEAD
        2e45846285c6afc396a7bbadaa9dad54360ed51c refs/remotes/origin/main
        2e45846285c6afc396a7bbadaa9dad54360ed51c refs/remotes/origin/test123
        """
        res = list(
            set(
            filter(
                lambda x: x not in ["HEAD"],
                map(
                    lambda x: x.strip().split()[-1].split("/")[-1],
                    self.out(*(git + ["show-ref"])).splitlines(),
                ),
            )
            )
        )
        return res

    @property
    def dirty(self):
        files = []
        for modifier, path in self._parse_git_status():
            if str(path) == "gimera.yml":
                continue
            files.append(path)
        return bool(files)

    def simple_commit_all(self, msg="."):
        self.X(*(git + ["add", "."]))
        self.X(*(git + ["commit", "--allow-empty", "-am", msg]))

    @property
    def hex(self):
        return self.out(*(git + ["log", "-n", "1", "--pretty=%H"]))

    def checkout(self, ref, force=False):
        self.X(*(git + ["checkout", "-f" if force else None, ref]))



import tempfile
import inquirer
import click
import sys
from pathlib import Path
from .repo import Repo
from .tools import _get_main_repo
from .consts import gitcmd as git
from .tools import prepare_dir
from .config import Config
from .patches import _apply_patchfile
from .patches import _technically_make_patch
from .patches import _apply_patchfile


def _commit(repo, branch, message, preview):
    config = Config()
    repo = config.get_repos(Path(repo))
    path2 = Path(tempfile.mktemp(suffix="."))
    main_repo = _get_main_repo()
    assert len(repo) == 1
    repo = repo[0]

    with prepare_dir(path2) as path2:
        path2 = path2 / "repo"
        gitrepo = Repo(path2)
        main_repo.X(*(git + ["clone", repo.url, path2]))

        if not branch:
            res = click.confirm(
                f"\n\n\nCommitting to branch {repo.branch} - continue?", default=True
            )
            if not res:
                sys.exit(-1)
            branch = repo.branch

        gitrepo.X(*(git + ["checkout", "-f", branch]))
        src_path = main_repo.path / repo.path
        patch_content = _technically_make_patch(main_repo, src_path)

        patchfile = gitrepo.path / "1.patch"
        patchfile.write_text(patch_content)

        _apply_patchfile(patchfile, gitrepo.path, error_ok=False)

        patchfile.unlink()
        gitrepo.X(*(git + ["add", "."]))
        if preview:
            gitrepo.X(*(git + ["diff"]))
            doit = inquirer.confirm("Commit this?", default=True)
            if not doit:
                return
        gitrepo.X(*(git + ["commit", "--no-verify", "-m", message]))
        gitrepo.X(*(git + ["push"]))



import itertools
from .fixtures import *  # required for all
import shutil
import os
import yaml
import subprocess
from ..consts import gitcmd as git
from ..repo import Repo
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_apply
from ..tools import safe_relative_to
from ..tools import get_effective_state
from ..tools import _make_sure_hidden_gimera_dir

token = {"token": 0}

import inspect
import os
from pathlib import Path

current_file = Path(os.path.abspath(inspect.getfile(inspect.currentframe())))
basename = current_file.stem

combinations = [
    ("S", "S", "S"),
    ("S", "S", "I"),
    ("S", "I", "S"),
    ("I", "S", "S"),
    ("S", "I", "I"),
    ("I", "S", "I"),
    ("I", "I", "S"),
    ("I", "I", "I"),
]

if basename == "test_snapshots_complex":
    for combo in combinations:
        letters = "".join(combo)
        filename = f"{basename}_{letters}.py"
        path = current_file.parent / filename
        code = current_file.read_text()
        code = code.replace("__COMBO__ = [('I', 'S', 'I')]", f"__COMBO__ = [{combo}]")
        code = code.replace(
            "test_snapshot_complex_ISI", f"test_snapshot_complex_{letters}"
        )
        path.write_text(code)
__COMBO__ = [('I', 'S', 'I')]


def test_snapshot_complex_ISI(temppath):
    for I, combo in enumerate(__COMBO__):

        def _e(x):
            assert x in ("S", "I")
            return "submodule" if x == "S" else "integrated"

        type1 = _e(combo[0])
        type11 = _e(combo[1])
        type111 = _e(combo[2])

        _test_snapshot_and_restore_complex_add_delete_modify_direct_subrepo_submodule(
            temppath,
            type1,
            type11,
            type111,
        )
        shutil.rmtree(temppath / "sub1.1.1")
        shutil.rmtree(temppath / "sub1.1")
        shutil.rmtree(temppath / "sub1")
        shutil.rmtree(temppath / "mainrepo")
        shutil.rmtree(temppath / "test_snapshot_and_restore")


def _test_snapshot_and_restore_complex_add_delete_modify_direct_subrepo_submodule(
    temppath, type1, type11, type111
):
    repo_sub111 = _make_remote_repo(temppath / "sub1.1.1")
    repo_sub11 = _make_remote_repo(temppath / "sub1.1")
    repo_sub1 = _make_remote_repo(temppath / "sub1")
    with clone_and_commit(repo_sub111, "branch1") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_sub11, "branch1") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        (repopath / "gimera.yml").write_text(
            yaml.dump(
                {
                    "repos": [
                        {
                            "url": str(repo_sub111),
                            "branch": "branch1",
                            "path": "a111/b111/sub1.1.1",
                            "patches": [],
                            "type": type111,
                        }
                    ]
                }
            )
        )
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_sub1, "branch1") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        (repopath / "gimera.yml").write_text(
            yaml.dump(
                {
                    "repos": [
                        {
                            "url": str(repo_sub11),
                            "branch": "branch1",
                            "path": "a11/b11/sub1.1",
                            "patches": [],
                            "type": type11,
                        }
                    ]
                }
            )
        )
        Repo(repopath).simple_commit_all()

    repos_yaml = {
        "repos": [
            {
                "url": str(repo_sub1),
                "branch": "branch1",
                "path": "a1/b1/sub1",
                "patches": [],
                "type": type1,
            },
        ]
    }
    _test_snapshot_and_restore_simple_add_delete_modify_direct(
        temppath, repo_sub1, repo_sub11, repo_sub111, repos_yaml
    )


def _test_snapshot_and_restore_simple_add_delete_modify_direct(
    temppath, repo_sub1, repo_sub11, repo_sub111, repos_yaml
):
    from ..snapshot import snapshot_recursive
    from ..snapshot import snapshot_restore

    workspace = temppath / "test_snapshot_and_restore"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "mainrepo")

    subprocess.check_output(
        git + ["clone", str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    (workspace_main / "gimera.yml").write_text(yaml.dump(repos_yaml))
    (workspace_main / "main.txt").write_text("main repo")
    repo = Repo(workspace_main)
    _make_sure_hidden_gimera_dir(repo.path)
    repo.simple_commit_all()
    repo.X(*(git + ["push"]))

    # change every level of the repo for its own; then change all levels and check
    for mode in ["use_gimera_migrate", "direct_snapshots"]:
    # for mode in ["direct_snapshots"]:
    # for mode in ["use_gimera_migrate"]:
        for i, adapted_paths in enumerate(
            [
                ["a1/b1/sub1"],
                ["a1/b1/sub1/a11/b11/sub1.1"],
                ["a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1"],
                [
                    "a1/b1/sub1/a11/b11/sub1.1",
                    "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                ],
                [
                    "a1/b1/sub1",
                    "a1/b1/sub1/a11/b11/sub1.1",
                    "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                ],
                [
                    "a1/b1/sub1/a11/b11/sub1.1",
                    "a1/b1/sub1",
                    "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                ],
                [
                    "a1/b1/sub1/a11/b11/sub1.1",
                    "a1/b1/sub1",
                    "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                ],
                [
                    "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                    "a1/b1/sub1/a11/b11/sub1.1",
                    "a1/b1/sub1",
                ],
                [
                    "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                    "a1/b1/sub1",
                    "a1/b1/sub1/a11/b11/sub1.1",
                ],
                ["a1/b1/sub1", "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1"],
                ["a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1", "a1/b1/sub1"],
            ]
        ):
            if workspace_main.exists():
                shutil.rmtree(workspace_main)
            subprocess.check_output(
                git + ["clone", str(repo_main), workspace_main],
                cwd=workspace.parent,
            )
            os.chdir(workspace_main)
            gimera_apply([], None, recursive=True, strict=True)
            # assert everything is there
            assert (
                workspace_main
                / "a1/b1/sub1"
                / "a11/b11/sub1.1"
                / "a111/b111/sub1.1.1"
                / "repo_sub.txt"
            ).exists()

            for adapted_path in adapted_paths:
                # make it dirty
                dirty_file = workspace_main / adapted_path / "file1.txt"
                original_content = dirty_file.read_text()
                dirty_file.write_text("i changed the file")
                # add a file
                added_file = workspace_main / adapted_path / "newfile.txt"
                added_file.write_text("new file")
                # delete a file
                deleted_file = workspace_main / adapted_path / "dont_look_at_me"
                deleted_file.unlink()

            os.chdir(workspace_main)
            token["token"] += 1
            os.environ["GIMERA_TOKEN"] = str(token["token"])
            if mode == "direct_snapshots":
                snapshot_path = workspace_main / "a1/b1/sub1"
                snapshot_recursive(workspace_main, [snapshot_path])

            # reapply
            os.chdir(workspace_main)

            if mode == "direct_snapshots":
                os.environ["GIMERA_FORCE"] = "1"
                gimera_apply([], None, recursive=True, strict=True)

                for adapted_path in adapted_paths:
                    dirty_file = workspace_main / adapted_path / "file1.txt"
                    assert dirty_file.read_text() == original_content
            else:
                os.environ["GIMERA_FORCE"] = "0"
                os.environ["PYTHONBREAKPOINT"] = "pudb"
                gimera_apply(
                    [], None, recursive=True, migrate_changes=True, strict=True
                )

            # restore
            if mode == "direct_snapshots":
                snapshot_restore(workspace_main, [snapshot_path])

            for adapted_path in adapted_paths:
                _assure_kept_changes(workspace_main, adapted_path)

            if mode == "use_gimera_migrate":
                # switch to other mode integrated/submodule and check
                # if changes are transported
                for j, adapted_path in enumerate(adapted_paths):
                    state = get_effective_state(
                        workspace_main, workspace_main / adapted_path, {}
                    )
                    closest_gimera = state["closest_gimera"]
                    parent_gimera = state["parent_gimera"]
                    parent_repo = Repo(state["parent_repo"])
                    working_dir = parent_gimera
                    relpath = safe_relative_to(
                        workspace_main / adapted_path, working_dir
                    )
                    effstate = get_effective_state(
                        workspace_main, workspace_main / adapted_path, {}
                    )
                    if effstate["is_submodule"]:
                        other_type = "integrated"
                    else:
                        other_type = "submodule"
                    pwd = os.getcwd()
                    os.chdir(working_dir)
                    token["token"] += 1
                    os.environ["GIMERA_TOKEN"] = str(token["token"])

                    def _apply():
                        gimera_apply(
                            [effstate["parent_gimera_relpath"]],
                            None,
                            force_type=other_type,
                            recursive=True,
                            migrate_changes=True,
                            strict=True,
                        )

                    # Explanation for gimera force = True:
                    # If submodules receive commits they get changed; so
                    # they appear as "new commits" in git status. It is uncertain,
                    # that other commits were done to the sup repo as well and then
                    # just deleting the subrepo would be a loss.
                    # Keeping this comment: we check transferring the changes, so just
                    # forcing takes some of the test complexity away.
                    # We explicitly commit the dirty sub repos.
                    assert os.getenv("GIMERA_FORCE") == "0"

                    for path in [
                        ".",
                        "a1/b1/sub1",
                        "a1/b1/sub1/a11/b11/sub1.1",
                        "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                        ".",
                        "a1/b1/sub1",
                        "a1/b1/sub1/a11/b11/sub1.1",
                        "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                        ".",
                        "a1/b1/sub1",
                        "a1/b1/sub1/a11/b11/sub1.1",
                        "a1/b1/sub1/a11/b11/sub1.1/a111/b111/sub1.1.1",
                    ]:
                        repo = Repo(workspace_main / path)
                        list(repo.get_submodules_with_new_commits())
                        for submodule in list(repo.get_submodules_with_new_commits()):
                            repo.X(*(git + ["add", submodule.path]))
                            repo.X(*(git + ["commit", "-m", "committed changes"]))
                    _apply()
                    os.chdir(pwd)
                    _assure_kept_changes(workspace_main, adapted_path)


def _assure_kept_changes(workspace_main, adapted_path):
    dirty_file = workspace_main / adapted_path / "file1.txt"
    added_file = workspace_main / adapted_path / "newfile.txt"
    deleted_file = workspace_main / adapted_path / "dont_look_at_me"
    # make sure that situation is like before:
    state = get_effective_state(workspace_main, workspace_main / adapted_path, {})
    parent_repo = state["parent_repo"]

    assert dirty_file.read_text() == "i changed the file"
    assert added_file.exists()
    assert not deleted_file.exists()

    os.environ["BREAKPOINT"] = "1"
    for file in [dirty_file, added_file, deleted_file]:
        if state["is_submodule"]:
            repo = Repo(adapted_path)
        else:
            repo = Repo(parent_repo)
        assert (
            file in repo.all_dirty_files_absolute
        ), f"{file} not in {repo.all_dirty_files}"



from .fixtures import *  # required for all
import itertools
import yaml
from contextlib import contextmanager
from ..repo import Repo
import os
import subprocess
import inspect
import os
from pathlib import Path
from .tools import gimera_apply
from ..tools import rsync
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_commit

from ..consts import gitcmd as git
import time

current_dir = Path(
    os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
)


def test_snapshot_switch_around_and_check_if_everything_is_there_several_subpaths(
    temppath,
):
    _test_snapshot_switch_around_and_check_if_everything_is_there(temppath, "a/b/sub1")


def test_snapshot_switch_around_and_check_if_everything_is_there_direct_root(temppath):
    _test_snapshot_switch_around_and_check_if_everything_is_there(temppath, "sub1")


def _test_snapshot_switch_around_and_check_if_everything_is_there(temppath, sub_path):
    """
    Challenge is, that the logic of snapshots can handle the root paths
    """
    workspace = (
        temppath / "test_snapshot_switch_around_and_check_if_everything_is_there"
    )
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = _make_remote_repo(temppath / "main")
    repo_sub1 = _make_remote_repo(temppath / "sub1")

    with clone_and_commit(repo_sub1, "branch1") as repopath:
        (repopath / "repo_sub.txt").write_text("This is a new function")
        Repo(repopath).simple_commit_all()

    with clone_and_commit(repo_main, "main") as repopath:
        (repopath / "repo_main.txt").write_text("This is a new function")
        (repopath / "dont_look_at_me").write_text("i am ugly")
        (repopath / "gimera.yml").write_text(
            yaml.dump(
                {
                    "repos": [
                        {
                            "url": str(repo_sub1),
                            "branch": "branch1",
                            "path": sub_path,
                            "type": "integrated",
                        },
                    ]
                }
            )
        )
        Repo(repopath).simple_commit_all()

    subprocess.check_output(
        git + ["clone", "file://" + str(repo_main), workspace_main],
        cwd=workspace.parent,
    )
    os.chdir(workspace_main)
    gimera_apply([], None)
    # assert everything is there
    assert (workspace_main / sub_path).exists()

    # make it dirty
    dirty_file1 = workspace_main / sub_path / "repo_sub.txt"
    assert dirty_file1.exists()
    dirty_file1.write_text("1")
    os.environ["GIMERA_FORCE"] = "0"
    gimera_apply([], None, recursive=True, migrate_changes=True, strict=True)

    assert dirty_file1.read_text() == "1"
    gimera_apply(
        [],
        None,
        force_type="submodule",
        recursive=True,
        migrate_changes=True,
        strict=True,
    )

    assert dirty_file1.read_text() == "1"
    gimera_apply(
        [],
        None,
        force_type="integrated",
        recursive=True,
        migrate_changes=True,
        strict=True,
    )



from itertools import islice
import subprocess
import yaml
import tempfile
import time
from datetime import datetime
import shutil
import uuid
import os
from pathlib import Path
import click
import sys
from curses import wrapper
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from . import runtime_state



def is_forced():
    return os.getenv("GIMERA_FORCE", "0") == "1"


def yieldlist(method):
    def wrapper(*args, **kwargs):
        result = list(method(*args, **kwargs))
        return result

    return wrapper


def X(*params, output=False, cwd=None, allow_error=False, env=None):
    """
    Catching output error and stderr
    try:
        stdout = X(output=True, .....)
    except subprocess.CalledProcessError as ex:
        stderr = ex.stderr
    """
    params = list(filter(lambda x: x is not None, list(params)))
    env2 = {k: v for k, v in os.environ.items()}
    env2.update(env or {})
    if params and params[0] == 'git' and os.getenv("GIMERA_QUIET") == "1":
        if any(x in params for x in ["commit", "push"]):
            params.append("--quiet")
    if output:
        ret = subprocess.run(
            params,
            encoding="utf-8",
            cwd=cwd,
            capture_output=True,
            text=True,
            env=env2,
        )
        if ret.returncode:
            if allow_error:
                return ""
            raise subprocess.CalledProcessError(
                returncode=ret.returncode,
                cmd=params,
                output=ret.stdout,
                stderr=ret.stderr,
            )
        return ret.stdout.rstrip()
    try:
        return subprocess.check_call(params, cwd=cwd, env=env2)
    except subprocess.CalledProcessError:
        if allow_error:
            return None
        raise


def _raise_error(msg):
    click.secho(msg, fg="red")
    if os.getenv("GIMERA_EXCEPTION_THAN_SYSEXIT") == "1":
        raise Exception(msg)
    else:
        sys.exit(-1)


def _strip_paths(paths):
    for x in paths:
        yield str(Path(x))


def safe_relative_to(path, path2):
    try:
        res = Path(path).relative_to(path2)
    except ValueError:
        return False
    else:
        return res


def is_empty_dir(path):
    return not any(Path(path).rglob("*"))


@contextmanager
def prepare_dir(path):
    tmp_path = path.parent / f"{path.name}.{uuid.uuid4()}"
    assert path.parent.exists()
    assert len(path.parts) > 1
    tmp_path.mkdir(parents=True)
    try:
        yield tmp_path
        if path.exists():
            rmtree(path)
        shutil.move(tmp_path, path)
    except Exception as ex:
        raise
    finally:
        if tmp_path.exists():
            try:
                rmtree(tmp_path)
            except Exception:
                pass


def file_age(path):
    try:
        mtime = datetime.fromtimestamp(os.stat(path).st_mtime)
    except:
        return 0
    return (datetime.now() - mtime).total_seconds()


@contextmanager
def wait_git_lock(path):
    MAX_TIMEOUT = 3600
    from .filelock import FileLock

    gimera_lock = path / ".git" / "gimera.lock"
    if not gimera_lock.parent.exists():
        gimera_lock = None
    index_lock = path / ".git" / "index.lock"

    def lock_exists():
        if index_lock.exists():
            return True
        if gimera_lock and gimera_lock.exists():
            return True
        return False

    if not index_lock.exists():
        yield
    else:
        while lock_exists():
            if file_age(index_lock) > MAX_TIMEOUT:
                index_lock.unlink()
                continue

            if gimera_lock and file_age(gimera_lock) > MAX_TIMEOUT:
                gimera_lock.unlink()
                continue

            time.sleep(0.5)

        if gimera_lock:
            with FileLock(gimera_lock, timeout=MAX_TIMEOUT):
                yield
        else:
            yield


def rmtree(path):
    try:
        shutil.rmtree(path)
    except:
        click.secho(f"Failed to remove {path}", fg="red")
        sys.exit(-1)


@contextmanager
def remember_cwd(cwd):
    old = os.getcwd()
    if cwd is not None:
        os.chdir(cwd)
    try:
        yield Path(cwd)
    finally:
        os.chdir(old)


def confirm(msg, raise_exception=True):
    if os.getenv("GIMERA_NON_INTERACTIVE") == "1":
        return True
    click.secho(msg, fg="yellow")
    res = click.confirm("Continue?", default=True)
    if not res and raise_exception:
        _raise_error("Aborted by user")
    return res


def retry(func, attempts=3, sleep=5):
    for i in range(attempts):
        try:
            func()
        except Exception:
            if i == attempts - 1:
                raise
            time.sleep(sleep)
        else:
            break


def try_rm_tree(path):
    if not path.exists():
        return

    def _action():
        shutil.rmtree(path)

    retry(func=_action)


@contextmanager
def temppath(mkdir=True, reuse_key=None):
    """
    reuse_key: keeps the directory - for reusability while gimera runs
    """
    if reuse_key in runtime_state['temppaths']:
        yield runtime_state['temppaths'][reuse_key]
        return

    with TemporaryDirectory() as path:
        path = Path(path)

    try:
        if mkdir:
            path.mkdir()
        yield path
    finally:
        if reuse_key is None:
            try_rm_tree(path)
        else:
            runtime_state['temppaths'][reuse_key] = path


def path1inpath2(path1, path2):
    try:
        path1.relative_to(path2)
        return True
    except Exception:
        return False


def rsync(dir1, dir2, exclude=None, delete_after=True):
    cmd = [
        "rsync",
        "-ar",
    ]
    if os.getenv("GIMERA_QUIET") != "1":
        cmd += [
            "--info=progress2",
        ]
    if delete_after:
        cmd += ["--delete-after"]
    for X in exclude or []:
        cmd += [f"--exclude={X}"]
    cmd.append(str(dir1) + "/")
    cmd.append(str(dir2) + "/")
    subprocess.check_call(cmd)


def get_url_type(url):
    if url.startswith("http"):
        return "http"
    if url.startswith("git@"):
        return "git"
    if url.startswith("/") or url.startswith("file://"):
        return "file"
    raise NotImplementedError(url)


def reformat_url(url, ttype):
    assert ttype in ["http", "git"]

    current = get_url_type(url)
    if ttype == current:
        return url
    elif ttype == "http" and current == "git":
        if url.startswith("git@"):
            url = url[4:]
        if ":" in url:
            # colon in password possible
            url = "/".join(url.split(":", 1))
        url = "https://" + url
        return url
    elif ttype == "git" and current == "http":
        url = url.split("://", 1)[1]
        url = f"git@{url}"
        url = ":".join(url.split("/", 1))
        return url
    else:
        raise NotImplementedError(f"{url} + {ttype}")


def verbose(txt):
    if not os.getenv("GIMERA_VERBOSE") == "1":
        return
    click.secho(txt, fg="yellow")


def get_nearest_repo(end, start):
    from .repo import Repo

    relpath = safe_relative_to(start, end)
    result = {"repo": Repo(end)}

    def walk(parent_module, path):
        result["repo"] = parent_module
        for submodule in parent_module.get_submodules():
            submodulepath = path / safe_relative_to(submodule.path, parent_module.path)
            if str(relpath).startswith(str(submodulepath)):
                walk(submodule, submodulepath)
                break

    walk(result["repo"], path=Path("."))
    return result["repo"].path


def _make_sure_hidden_gimera_dir(root_dir):
    from .repo import Repo
    from .consts import gitcmd as git

    path = Path(root_dir) / ".gitignore"
    if not path.exists():
        path.write_text(".gimera\n")
    else:
        content = path.read_text().splitlines()
        if not [x for x in content if x == ".gimera"]:
            content.append(".gimera")
            path.write_text("\n".join(content))
            repo = Repo(root_dir)
            if repo.staged_files:
                _raise_error("No staged files allowed, when changing .gitignore")
            repo.X(*(git + ["add", ".gitignore"]))
            if repo.all_dirty_files:
                repo.X(*(git + ["commit", "--no-verify", "-m", "add .gimera to .gitignore"]))
    return root_dir / ".gimera"


@yieldlist
def _get_remotes(repo_yml):
    from .repo import Remote

    config = repo_yml.remotes
    if not config:
        return

    for name, url in dict(config).items():
        yield Remote(None, name, url)


def _get_main_repo():
    from .repo import Repo

    path = Path(os.getcwd())
    while True:
        if (path / ".git").exists() and (path / ".git").is_dir():
            break
        path = path.parent
        if len(path.parts) == 1:
            path = Path(os.getcwd())
            break

    return Repo(path)


def _get_missing_repos(config):
    for repo in config.repos:
        if not repo.enabled:
            continue
        if not repo.path.exists():
            yield repo


def get_closest_gimera(end, start):
    if start.is_dir():
        p = start
    else:
        p = start.parent
    while safe_relative_to(p, end):
        if (p / "gimera.yml").exists():
            return p
        p = p.parent
    return end
    raise Exception(f"No parent gimera found for {start}")


def collect_upwards_common_vars(root_dir, start_gimera):
    result = {}

    def x(content):
        content = yaml.load(content, Loader=yaml.FullLoader)
        for k, v in content.get("common", {}).get("vars", {}).items():
            if k not in result:
                result[k] = v

    file = start_gimera / "gimera.yml"
    while file:
        x(file.read_text())
        file = get_closest_gimera(root_dir, file)
        if file:
            file = file / "gimera.yml"
    return result


def get_effective_state(root_dir, path, common_vars):
    from .repo import Repo

    path = Path(path)
    root_dir = Path(root_dir)

    # closest_gimera = local_gimera
    from .config import Config

    closest_gimera = get_closest_gimera(root_dir, path / "dummy") or root_dir
    config = Config(
        force_gimera_file=closest_gimera / "gimera.yml",
        common_vars=common_vars,
    )
    for repo in config.repos:
        if repo.path == safe_relative_to(path, closest_gimera):
            path_is_provided_by_gimera_but_itself_no_gimera = True
            break
    else:
        path_is_provided_by_gimera_but_itself_no_gimera = False

    if path_is_provided_by_gimera_but_itself_no_gimera:
        parent_gimera = closest_gimera
    else:
        parent_gimera = get_closest_gimera(root_dir, closest_gimera.parent)
    if parent_gimera == root_dir:
        parent_repo = root_dir
    else:
        parent_repo = get_nearest_repo(root_dir, parent_gimera)

    parent_repo_relpath = safe_relative_to(path, parent_repo)
    parent_gimera_relpath = safe_relative_to(path, parent_gimera)
    is_submodule = Repo(parent_repo).is_path_a_submodule(parent_repo_relpath)
    return {
        "is_submodule": is_submodule,
        "closest_gimera": closest_gimera,
        "parent_gimera": parent_gimera,
        "parent_repo": parent_repo,
        "parent_gimera_relpath": parent_gimera_relpath,
        "parent_repo_relpath": parent_repo_relpath,
    }


def filter_files_to_folders(files, folders):
    for file in files:
        for folder in folders:
            if str(file).startswith(str(folder)):
                yield file


def files_relative_to(files, folder):
    for file in files:
        res = safe_relative_to(file, folder)
        if res:
            yield res

def replace_dir_with(source_dir, dest_dir):
    tmppath = Path(str(dest_dir) + "." + str(uuid.uuid4()))
    if dest_dir.exists():
        shutil.move(dest_dir, tmppath)

    try:
        shutil.move(source_dir, dest_dir)
    finally:
        if tmppath.exists():
            rmtree(tmppath)

def split_every(n, iterable, piece_maker=tuple):
    # thanks odoo
    iterator = iter(iterable)
    piece = piece_maker(islice(iterator, n))
    while piece:
        yield piece
        piece = piece_maker(islice(iterator, n))

@contextmanager
def assert_exception_no_exit():
    old_val = os.getenv("GIMERA_EXCEPTION_THAN_SYSEXIT", "0")
    try:
        os.environ["GIMERA_EXCEPTION_THAN_SYSEXIT"] = "1"
        yield
    finally:
        os.environ["GIMERA_EXCEPTION_THAN_SYSEXIT"] = old_val


from .fixtures import *  # required for all
import itertools
import yaml
from ..repo import Repo
import os
import subprocess
import shutil
import os
from pathlib import Path
from .tools import gimera_apply
from . import temppath
from .tools import _make_remote_repo
from .tools import clone_and_commit
from .tools import gimera_commit

from ..consts import gitcmd as git


def test_recursive_gimeras_2_levels(temppath):
    workspace = temppath / "workspace_tree_dirty_files"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = None
    repo_sub = None
    repo_subsub = None

    gimera_main = None
    gimera_sub = None

    # region prepare repos
    def prepare_repos(ttype_sub, ttype_subsub):
        if workspace_main.exists():
            shutil.rmtree(workspace_main)
        nonlocal repo_main, repo_sub, repo_subsub, gimera_main, gimera_sub
        if repo_main:
            shutil.rmtree(repo_main)
        if repo_sub:
            shutil.rmtree(repo_sub)
        if repo_subsub:
            shutil.rmtree(repo_subsub)
        repo_main = _make_remote_repo(temppath / "mainrepo")
        repo_sub = _make_remote_repo(temppath / "sub1")
        repo_subsub = _make_remote_repo(temppath / "subsub1")
        path = Path(os.environ["GIMERA_CACHE_DIR"])
        if path.exists():
            shutil.rmtree(path)
        path.mkdir(parents=True, exist_ok=True)

        # region gimera config
        gimera_main = {
            "repos": [
                {
                    "url": f"file://{repo_sub}",
                    "branch": "main",
                    "path": "sub",
                    "patches": [],
                    "type": ttype_sub,
                },
            ]
        }
        gimera_sub = {
            "repos": [
                {
                    "url": f"file://{repo_subsub}",
                    "branch": "main",
                    "path": "subsub",
                    "patches": [],
                    "type": ttype_subsub,
                },
            ]
        }
        # endregion

        with clone_and_commit(repo_main, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_main))
            (repopath / "main.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_sub))
            (repopath / "sub.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_subsub, "main") as repopath:
            (repopath / "subsub.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        subprocess.check_output(
            git + ["clone", "file://" + str(repo_main), workspace_main],
            cwd=workspace,
        )

    # endregion

    repo = Repo(workspace_main)

    # region: case 1: all integrated
    prepare_repos("integrated", "integrated")
    os.chdir(workspace_main)
    gimera_apply([], None, recursive=True, strict=True)

    assert (workspace_main / "sub" / "sub.txt").exists()
    assert (workspace_main / "sub" / "gimera.yml").exists()
    assert (workspace_main / "sub" / "subsub" / "subsub.txt").exists()
    dirty = [x for x in repo.all_dirty_files if x.stem != '.gitignore']
    assert not dirty
    # endregion

    # region: case 2: submodule then integrated
    prepare_repos("submodule", "integrated")
    os.chdir(workspace_main)
    gimera_apply([], update=None, recursive=True, strict=True)

    assert (workspace_main / "sub" / "sub.txt").exists()
    assert (workspace_main / "sub" / "gimera.yml").exists()
    assert (workspace_main / "sub" / "subsub" / "subsub.txt").exists()
    dirty = [x for x in repo.all_dirty_files if x.stem != '.gitignore']
    assert not dirty
    # endregion

    # region: case 3: integrated then submodule
    prepare_repos("integrated", "submodule")
    os.chdir(workspace_main)
    gimera_apply([], update=None, recursive=True, strict=True)

    assert (workspace_main / "sub" / "sub.txt").exists()
    assert (workspace_main / "sub" / "gimera.yml").exists()
    assert (workspace_main / "sub" / "subsub" / "subsub.txt").exists()
    dirty = [x for x in repo.all_dirty_files if x.stem != '.gitignore']
    assert not dirty
    # endregion

    # region: case 4: submodule submodule
    prepare_repos("submodule", "submodule")
    os.chdir(workspace_main)
    gimera_apply([], update=None, recursive=True, strict=True)

    assert (workspace_main / "sub" / "sub.txt").exists()
    assert (workspace_main / "sub" / "gimera.yml").exists()
    assert (workspace_main / "sub" / "subsub" / "subsub.txt").exists()
    dirty = [x for x in repo.all_dirty_files if x.stem != '.gitignore']
    assert not dirty
    # endregion


def test_recursive_gimeras_3_levels(temppath):
    workspace = temppath / "workspace_tree_dirty_files_3"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = None
    repo_sub1 = None
    repo_sub2 = None
    repo_sub3 = None

    gimera_main = None
    gimera_sub = None

    # region prepare repos
    def prepare_repos(ttype_sub1, ttype_sub2, ttype_sub3):
        if workspace_main.exists():
            shutil.rmtree(workspace_main)
        nonlocal repo_main, repo_sub1, repo_sub2, repo_sub3, gimera_main, gimera_sub
        if repo_main:
            shutil.rmtree(repo_main)
        if repo_sub1:
            shutil.rmtree(repo_sub1)
        if repo_sub2:
            shutil.rmtree(repo_sub2)
        if repo_sub3:
            shutil.rmtree(repo_sub3)
        repo_main = _make_remote_repo(temppath / "mainrepo")
        repo_sub1 = _make_remote_repo(temppath / "sub1")
        repo_sub2 = _make_remote_repo(temppath / "sub2")
        repo_sub3 = _make_remote_repo(temppath / "sub3")
        path = Path(os.environ["GIMERA_CACHE_DIR"])
        if path.exists():
            shutil.rmtree(path)
        path.mkdir(parents=True, exist_ok=True)

        # region gimera config
        gimera_main = {
            "repos": [
                {
                    "url": f"file://{repo_sub1}",
                    "branch": "main",
                    "path": "sub1",
                    "patches": [],
                    "type": ttype_sub1,
                },
            ]
        }
        gimera_sub = {
            "repos": [
                {
                    "url": f"file://{repo_sub2}",
                    "branch": "main",
                    "path": "sub2",
                    "patches": [],
                    "type": ttype_sub2,
                },
            ]
        }
        gimera_sub2 = {
            "repos": [
                {
                    "url": f"file://{repo_sub3}",
                    "branch": "main",
                    "path": "sub3",
                    "patches": [],
                    "type": ttype_sub3,
                },
            ]
        }
        # endregion

        with clone_and_commit(repo_main, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_main))
            (repopath / "main.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub1, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_sub))
            (repopath / "sub1.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub2, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_sub2))
            (repopath / "sub2.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub3, "main") as repopath:
            (repopath / "sub3.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        subprocess.check_output(
            git + ["clone", "file://" + str(repo_main), workspace_main],
            cwd=workspace,
        )

    # endregion

    permutations = list(sorted(set(itertools.permutations("000111", 3))))

    for i, permutation in enumerate(permutations):

        def ttype(x):
            return "integrated" if int(x) else "submodule"

        prepare_repos(*tuple(map(ttype, permutation)))
        os.chdir(workspace_main)
        gimera_apply([], update=None, recursive=True, strict=True)

        repo = Repo(workspace_main)
        assert (workspace_main / "sub1" / "sub1.txt").exists()
        assert (workspace_main / "sub1" / "gimera.yml").exists()
        assert (workspace_main / "sub1" / "sub2" / "sub2.txt").exists()
        assert (workspace_main / "sub1" / "sub2" / "sub3" / "sub3.txt").exists()
        dirty = [x for x in repo.all_dirty_files if x.stem != '.gitignore']
        assert not dirty


def test_recursive_gimeras_5_levels(temppath):
    workspace = temppath / "test_recursive_gimeras_4_levels"
    workspace.mkdir()
    workspace_main = workspace / "main_working"

    repo_main = None
    repo_sub1 = None
    repo_sub2 = None
    repo_sub3 = None
    repo_sub4 = None
    repo_sub5 = None

    gimera_main = None
    gimera_sub = None

    # region prepare repos
    def prepare_repos(ttype_sub1, ttype_sub2, ttype_sub3, ttype_sub4, ttype_sub5):
        if workspace_main.exists():
            shutil.rmtree(workspace_main)
        nonlocal repo_main, repo_sub1, repo_sub2, repo_sub3, repo_sub4, repo_sub5, gimera_main, gimera_sub
        if repo_main:
            shutil.rmtree(repo_main)
        if repo_sub1:
            shutil.rmtree(repo_sub1)
        if repo_sub2:
            shutil.rmtree(repo_sub2)
        if repo_sub3:
            shutil.rmtree(repo_sub3)
        if repo_sub4:
            shutil.rmtree(repo_sub4)
        if repo_sub5:
            shutil.rmtree(repo_sub5)
        repo_main = _make_remote_repo(temppath / "mainrepo")
        repo_sub1 = _make_remote_repo(temppath / "sub1")
        repo_sub2 = _make_remote_repo(temppath / "sub2")
        repo_sub3 = _make_remote_repo(temppath / "sub3")
        repo_sub4 = _make_remote_repo(temppath / "sub4")
        repo_sub5 = _make_remote_repo(temppath / "sub5")
        path = Path(os.environ["GIMERA_CACHE_DIR"])
        if path.exists():
            shutil.rmtree(path)
        path.mkdir(parents=True, exist_ok=True)

        # region gimera config
        gimera_main = {
            "repos": [
                {
                    "url": f"file://{repo_sub1}",
                    "branch": "main",
                    "path": "sub1",
                    "patches": [],
                    "type": ttype_sub1,
                },
            ]
        }
        gimera_sub = {
            "repos": [
                {
                    "url": f"file://{repo_sub2}",
                    "branch": "main",
                    "path": "sub2",
                    "patches": [],
                    "type": ttype_sub2,
                },
            ]
        }
        gimera_sub2 = {
            "repos": [
                {
                    "url": f"file://{repo_sub3}",
                    "branch": "main",
                    "path": "sub3",
                    "patches": [],
                    "type": ttype_sub3,
                },
            ]
        }
        gimera_sub3 = {
            "repos": [
                {
                    "url": f"file://{repo_sub4}",
                    "branch": "main",
                    "path": "sub4",
                    "patches": [],
                    "type": ttype_sub4,
                },
            ]
        }
        gimera_sub4 = {
            "repos": [
                {
                    "url": f"file://{repo_sub5}",
                    "branch": "main",
                    "path": "sub5",
                    "patches": [],
                    "type": ttype_sub5,
                },
            ]
        }
        # endregion

        with clone_and_commit(repo_main, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_main))
            (repopath / "main.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub1, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_sub))
            (repopath / "sub1.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub2, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_sub2))
            (repopath / "sub2.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub3, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_sub3))
            (repopath / "sub3.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub4, "main") as repopath:
            (repopath / "gimera.yml").write_text(yaml.dump(gimera_sub4))
            (repopath / "sub4.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        with clone_and_commit(repo_sub5, "main") as repopath:
            (repopath / "sub5.txt").write_text("This is a new function")
            Repo(repopath).simple_commit_all()

        subprocess.check_output(
            git + ["clone", "file://" + str(repo_main), workspace_main],
            cwd=workspace,
        )

    # endregion

    permutations = list(sorted(set(itertools.permutations("0000011111", 5))))

    for i, permutation in enumerate(permutations):

        def ttype(x):
            return "integrated" if int(x) else "submodule"

        prepare_repos(*tuple(map(ttype, permutation)))
        os.chdir(workspace_main)
        gimera_apply([], update=None, recursive=True, strict=True)

        repo = Repo(workspace_main)
        assert (workspace_main / "sub1" / "sub1.txt").exists()
        assert (workspace_main / "sub1" / "gimera.yml").exists()
        assert (workspace_main / "sub1" / "sub2" / "sub2.txt").exists()
        assert (workspace_main / "sub1" / "sub2" / "sub3" / "sub3.txt").exists()
        assert (
            workspace_main / "sub1" / "sub2" / "sub3" / "sub4" / "sub4.txt"
        ).exists()
        assert (
            workspace_main / "sub1" / "sub2" / "sub3" / "sub4" / "sub5" / "sub5.txt"
        ).exists()
        dirty = [x for x in repo.all_dirty_files if x.stem != '.gitignore']
        assert not dirty



import os
import time
import uuid
import click
import shutil
import subprocess
from pathlib import Path
from .consts import gitcmd as git
from .repo import Repo
from .tools import prepare_dir
from .tools import remember_cwd
from .tools import reformat_url
from .tools import _raise_error
from .tools import rmtree
from .tools import replace_dir_with
from .tools import temppath

# store big repos in tar file and try to restore from there;
# otherwise lot of downloads have to be done


from contextlib import contextmanager


def _make_cache_path(url):
    try:
        urlsafe = reformat_url(url, "git")
    except Exception:
        urlsafe = url
    for c in "?:+[]{}\\/\"'_":
        urlsafe = urlsafe.replace(c, "-")
    urlsafe = urlsafe.split("@")[-1]
    base = Path(os.environ.get("GIMERA_CACHE_DIR") or os.path.expanduser("~/.cache/gimera"))
    return base / urlsafe


def _invalidate_cache_if_needed(golden_path):
    must_exist = ["HEAD", "refs", "objects", "config"]
    if golden_path.exists() and (any(
        not (golden_path / x).exists() for x in must_exist
    ) or os.getenv("GIMERA_CLEAR_CACHE") == "1"):
        click.secho(f"Removing cache directory:\n{golden_path}", fg="red")
        rmtree(golden_path)

    if os.getenv("GIMERA_CLEAR_ZIP_CACHE", "") == "1":
        tar = _get_cache_dir_tarfile(golden_path)
        if tar.exists():
            click.secho(f"Removing cache tar file:\n{tar}", fg="red")
            tar.unlink()


def _clone_or_restore(main_repo, url, golden_path, possible_temp_path):
    click.secho(
        f"Caching the repository {url} for quicker reuse",
        fg="yellow",
    )
    tar = _get_cache_dir_tarfile(golden_path)
    with prepare_dir(possible_temp_path) as _path:
        with remember_cwd(
            "/tmp"
        ):  # called from other situations where path may not exist anymore
            restored = False
            if tar.exists():
                try:
                    _extract_tar_file(_path, tar)
                    restored = True
                except Exception:
                    click.secho(
                        f"Failed to extract tar file {tar} - will try to clone again.",
                        fg="red",
                    )
                    tar.unlink()

            if not restored:
                rmtree(_path)
                _path.mkdir(parents=True)
                Repo(main_repo.path).X(*(git + ["clone", "--bare", url, _path]))
                _make_tar_file(_path, tar)


def _ensure_sha(repo_yml, effective_path, update):
    if not repo_yml.sha:
        return
    repo = Repo(effective_path)
    if repo.contain_commit(repo_yml.sha):
        return
    repo.fetchall()
    if repo.contain_commit(repo_yml.sha):
        return
    if not update:
        _raise_error(
            (
                f"After fetching the commit {repo_yml.sha} "
                f"was not found for {repo_yml.path}.\n"
                f"All remote branches were checked."
            )
        )
    else:
        click.secho(
            f"Warning: commit {repo_yml.sha} not found "
            f"for {repo_yml.path} - will retry after update.",
            fg="yellow",
        )


@contextmanager
def _get_cache_dir(main_repo, repo_yml, no_action_if_not_exist=False, update=None):
    url = repo_yml.url
    if not url:
        _raise_error(f"Missing url for: {repo_yml.path}")

    golden_path = _make_cache_path(url)

    if os.getenv("GIMERA_NO_CACHE", "") == "1":
        TEMP_KEY = f"{repo_yml.url}_{repo_yml.sha or repo_yml.branch}"
        with temppath(mkdir=False, reuse_key=TEMP_KEY) as path:
            if not path.exists():
                subprocess.run(["git", "clone", "--single-branch", "--depth=1", "--branch", repo_yml.branch, repo_yml.url, path], check=True)
                if repo_yml.sha:
                    Repo(path).X(*(git + ["fetch", "origin", repo_yml.sha]))
                    Repo(path).X(*(git + ["checkout", repo_yml.sha]))
            yield path
            return

    if no_action_if_not_exist and not golden_path.exists():
        yield None
        return

    possible_temp_path = Path(str(golden_path) + "." + str(uuid.uuid4()))
    try:
        golden_path.parent.mkdir(exist_ok=True, parents=True)
        _invalidate_cache_if_needed(golden_path)

        just_cloned = False
        if not golden_path.exists():
            _clone_or_restore(main_repo, url, golden_path, possible_temp_path)
            just_cloned = True

        effective_path = possible_temp_path if just_cloned else golden_path
        _ensure_sha(repo_yml, effective_path, update)

        yield effective_path

        if just_cloned:
            replace_dir_with(possible_temp_path, golden_path)

    finally:
        possible_temp_path = Path(possible_temp_path)
        if possible_temp_path.exists():
            rmtree(possible_temp_path)


def _get_cache_dir_tarfile(_path):
    return Path(str(_path) + ".tar.gz")


def _make_tar_file(_path, tarfile):
    if tarfile.exists():
        tarfile.unlink()
    click.secho(
        f"Creating tar file {tarfile} from {_path} - might take some time on large repos.",
        fg="yellow",
    )
    tempfilename = str(tarfile) + "." + str(uuid.uuid4())
    subprocess.check_call(["tar", "cfz", str(tempfilename), "-C", str(_path), "."])
    os.replace(tempfilename, tarfile)


def _extract_tar_file(_path, tarfile):
    click.secho(f"Extracting tar file {tarfile} to {_path}", fg="yellow")
    subprocess.check_call(["tar", "xfz", str(tarfile)], cwd=_path)



import tempfile
from pathlib import Path
import os
import shutil
from ..consts import gitcmd as git
import subprocess
from contextlib import contextmanager


def gimera_apply(*args, **kwargs):
    from ..gimera import _apply

    return _apply(*args, **kwargs)


def gimera_commit(*args, **kwargs):
    from ..gimera import _commit

    return _commit(*args, **kwargs)


@contextmanager
def clone_and_commit(repopath, branch, checkout_options=None, commit=True):
    path = Path(tempfile.mktemp(suffix="."))
    if path.exists():
        shutil.rmtree(path)
    subprocess.check_call(git + ["clone", repopath, path], cwd=repopath)
    subprocess.check_call(
        git + ["checkout"] + (checkout_options or []) + [branch], cwd=path
    )
    try:
        yield path
        if commit:
            subprocess.check_call(
                git + ["push", "--set-upstream", "origin", branch], cwd=path
            )
    finally:
        shutil.rmtree(path)


def _make_remote_repo(path):
    path.mkdir(parents=True)
    subprocess.check_call(["git", "init", "--bare", "--initial-branch=main"], cwd=path)

    tmp = path.parent / "tmp"
    subprocess.check_call(git + ["clone", f"file://{path}", tmp])
    (tmp / "file1.txt").write_text("random repo on main")
    subprocess.check_call(git + ["add", "file1.txt"], cwd=tmp)
    subprocess.check_call(git + ["commit", "-am", "on main"], cwd=tmp)
    subprocess.check_call(git + ["push"], cwd=tmp)

    subprocess.check_call(git + ["checkout", "-b", "branch1"], cwd=tmp)
    (tmp / "file1.txt").write_text("random repo on branch1")
    subprocess.check_call(git + ["add", "file1.txt"], cwd=tmp)
    subprocess.check_call(git + ["commit", "-am", "on branch1"], cwd=tmp)
    subprocess.check_call(
        git + ["push", "--set-upstream", "origin", "branch1"], cwd=tmp
    )

    shutil.rmtree(tmp)
    return path



from contextlib import contextmanager
import os
from pathlib import Path
import click
from .repo import Repo
from .fetch import _fetch_repos_in_parallel
from .tools import _get_main_repo
from .tools import _raise_error, safe_relative_to
from .consts import gitcmd as git
from .consts import REPO_TYPE_INT, REPO_TYPE_SUB
from .config import Config
from .patches import make_patches
from .tools import verbose
from .integrated import _update_integrated_module
from .submodule import _make_sure_subrepo_is_checked_out
from .snapshot import snapshot_recursive, snapshot_restore
from .submodule import _fetch_latest_commit_in_submodule
from .submodule import __add_submodule
from .tools import get_closest_gimera
from .tools import get_effective_state
from .tools import _make_sure_hidden_gimera_dir
from .cachedir import _get_cache_dir


def _check_sha_belongs_to_branch(main_repo, repo_yml):
    if not repo_yml.sha:
        return

    with _get_cache_dir(main_repo, repo_yml, no_action_if_not_exist=True) as cache_dir:
        if not cache_dir:
            return
        repo = Repo(cache_dir)
        try:
            repo.X(*(git + ["merge-base", "--is-ancestor", repo_yml.sha, repo_yml.branch]))
        except Exception:
            if os.getenv("GIMERA_NON_INTERACTIVE") == "1":
                click.secho(
                    f"SHA {repo_yml.sha} does not belong to branch "
                    f"{repo_yml.branch}. Removing SHA.",
                    fg="yellow",
                )
            else:
                answer = click.prompt(
                    f"SHA {repo_yml.sha} does not belong to branch "
                    f"{repo_yml.branch}. Removing SHA? (ok or abort)",
                    type=click.Choice(["ok", "abort"], case_sensitive=False),
                )
                if answer == "abort":
                    _raise_error("Aborted by user.")
            repo_yml._sha = None


def _apply(
    repos,
    update,
    force_type=False,
    strict=False,
    recursive=False,
    no_patches=False,
    remove_invalid_branches=False,
    auto_commit=True,
    no_fetch=False,
    migrate_changes=False,
    raise_exception=False,
):
    """
    :param repos: user input parameter from commandline
    :param update: bool - flag from command line
    """
    if raise_exception:
        os.environ["GIMERA_EXCEPTION_THAN_SYSEXIT"] = "1"
    if migrate_changes:
        no_patches = True


    sub_path = None
    main_repo = _get_main_repo()
    closest_gimera = (
        get_closest_gimera(main_repo.path, Path(os.getcwd()) / "dummy")
        or main_repo.path
    )
    if not sub_path:
        _make_sure_hidden_gimera_dir(main_repo.path)
    os.chdir(closest_gimera)
    if main_repo.path != closest_gimera:
        sub_path = closest_gimera

    _internal_apply(
        repos,
        update,
        force_type,
        strict=strict,
        recursive=recursive,
        no_patches=no_patches,
        remove_invalid_branches=remove_invalid_branches,
        auto_commit=auto_commit,
        no_fetch=no_fetch,
        sub_path=sub_path,
        migrate_changes=migrate_changes,
    )


def _internal_apply(
    repos,
    update,
    force_type,
    strict=False,
    recursive=False,
    no_patches=False,
    common_vars=None,
    parent_config=None,
    auto_commit=True,
    sub_path=None,
    no_fetch=None,
    migrate_changes=None,
    **options,
):
    common_vars = common_vars or {}
    main_repo = _get_main_repo()
    config = Config(
        force_type=force_type,
        recursive=recursive,
        common_vars=common_vars,
        parent_config=parent_config,
    )
    repos = config.get_repos(repos)
    # update repos in parallel to be faster
    _fetch_repos_in_parallel(
        main_repo, repos, update=update, minimal_fetch=no_fetch, no_fetch=no_fetch
    )
    if sub_path:
        verbose(f"internal apply at sub path: {sub_path}")
    # does not work in sub repos, because at apply at this point in time
    # the files are not committed and still dirty
    common_vars.update(config.yaml_config.get("common", {}).get("vars", {}))
    verbose(f"common vars: {common_vars}")
    with main_repo.stay_at_commit(not auto_commit and not sub_path):
        if migrate_changes:
            relative_sub_path = (
                sub_path and safe_relative_to(sub_path, main_repo.path) or Path(".")
            )
            snapshot_recursive(
                main_repo.path,
                [main_repo.path / relative_sub_path / repo.path for repo in repos],
            )

        try:
            for repo in repos:
                verbose(f"applying {repo.path}")
                if not update:
                    _check_sha_belongs_to_branch(main_repo, repo)
                _turn_into_correct_repotype(
                    sub_path or main_repo.path,
                    main_repo,
                    repo,
                    config,
                    common_vars,
                )
                if repo.type == REPO_TYPE_SUB:
                    _make_sure_subrepo_is_checked_out(
                        sub_path or main_repo.path, main_repo, repo, common_vars
                    )
                    _fetch_latest_commit_in_submodule(
                        sub_path or main_repo.path,
                        main_repo,
                        repo,
                        common_vars,
                        update=update,
                    )
                elif repo.type == REPO_TYPE_INT:
                    if not no_patches:
                        make_patches(
                            sub_path or main_repo.path, main_repo, repo, common_vars
                        )

                    try:
                        _update_integrated_module(
                            sub_path or main_repo.path,
                            main_repo,
                            repo,
                            update,
                            common_vars,
                            **options,
                        )
                    except Exception as ex:
                        msg = (
                            f"Error updating integrated submodules for: {repo.path}\n\n{ex}"
                        )
                        _raise_error(msg)

                    if not strict:
                        # fatal: refusing to create/use '.git/modules/addons_connector/modules/addons_robot/aaa' in another submodule's git dir
                        # not submodules inside integrated modules
                        force_type = REPO_TYPE_INT

                if recursive:
                    _apply_subgimera(
                        main_repo,
                        repo,
                        update,
                        force_type if not strict else None,
                        strict=strict,
                        no_patches=no_patches,
                        common_vars=common_vars,
                        parent_config=config,
                        auto_commit=auto_commit,
                        sub_path=sub_path,
                        migrate_changes=False,  # already done
                        **options,
                    )

                    # if subgimera is a git submodule and was committed and changed,
                    # then this parent gimera is dirty; so we also commit the child
                    if auto_commit:
                        state = get_effective_state(
                            main_repo.path, (sub_path or main_repo.path) / repo.path, common_vars
                        )
                        # commit a gitmodule if sha updated
                        parent_repo = Repo(state["parent_repo"])
                        # Path(repo.path)
                        if state['parent_repo_relpath'] in parent_repo.all_dirty_files:
                            parent_repo.commit_dir_if_dirty(
                                state['parent_repo_relpath'], "gimera: updated submodule"
                            )
                        del parent_repo

                        # commit updated gimera if e.g. sha changed
                        parent_repo = Repo(state["parent_repo"])
                        parent_repo_relpath = state['parent_repo_relpath']
                        parent_repo_gimera = Path(parent_repo_relpath / 'gimera.yml')
                        if parent_repo_gimera in parent_repo.all_dirty_files:
                            parent_repo.commit_dir_if_dirty(
                                parent_repo_gimera, "gimera: updated submodule"
                            )
                        del parent_repo
        finally:
            if migrate_changes:
                snapshot_restore(
                    main_repo.path,
                    [main_repo.path / relative_sub_path / repo.path for repo in repos],
                )


def _apply_subgimera(
    main_repo,
    repo,
    update,
    force_type,
    strict,
    no_patches,
    parent_config,
    sub_path,
    common_vars,
    **options,
):
    subgimera = Path(repo.path) / "gimera.yml"
    if sub_path and sub_path.relative_to(main_repo.path) == Path("."):
        sub_path = main_repo.path

    new_sub_path = Path(sub_path or main_repo.path) / repo.path
    if not subgimera.exists():
        return
    pwd = os.getcwd()

    try:
        verbose(f"apply subgimera: {subgimera}")
        os.chdir(new_sub_path)
        _internal_apply(
            [],
            update,
            force_type=force_type,
            strict=strict,
            recursive=True,
            no_patches=no_patches,
            parent_config=parent_config,
            sub_path=new_sub_path,
            common_vars=common_vars,
            **options,
        )

        state = get_effective_state(main_repo.path, new_sub_path, common_vars)
        parent_repo = Repo(state["parent_repo"])

        dirty_files = list(
            filter(
                lambda x: safe_relative_to(x, new_sub_path), parent_repo.all_dirty_files
            )
        )
        if dirty_files:
            parent_repo.please_no_staged_files()
            for f in dirty_files:
                parent_repo.X(*(git + ["add", f]))
            parent_repo.X(
                *(git + ["commit", "--no-verify", "-m", f"gimera: updated sub path {repo.path}"])
            )
        # commit submodule updates or changed dirs
    finally:
        os.chdir(pwd)


def _turn_into_correct_repotype(
    working_dir, main_repo, repo_config, config, common_vars
):
    """
    if git submodule and exists: nothing todo
    if git submodule and not exists: cloned
    if git submodule and already exists a path: path removed, submodule added

    if integrated and exists no sub: nothing todo
    if integrated and not exists: cloned (later not here)
    if integrated and git submodule and already exists a path: submodule removed

    """
    verbose(f"turn into correct repotype: {repo_config.path}")
    state = get_effective_state(
        main_repo.path, working_dir / repo_config.path, common_vars
    )
    repo = Repo(state["parent_repo"])
    if repo_config.type == REPO_TYPE_INT:
        if state["is_submodule"]:
            # always delete
            repo.force_remove_submodule(state["parent_repo_relpath"])
    else:
        __add_submodule(
            main_repo.path, working_dir, repo, repo_config, config, common_vars
        )
